import parser._
import proto._

val pobj = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      @0 := @0 - n
    end
    def n()
      @0
    end
  end
end
let c = counter(4) in
  print(c.n());
  print(c.inc(4));
  print(c.dec(2))
end
""")
eval(pobj)
val pobjself = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      self.inc(-n)
    end
    def n()
      @0
    end
  end
end
let c = counter(4) in
  print(c.n());
  print(c.inc(4));
  print(c.dec(2))
end
""")
eval(pobjself)

val pobj2 = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      self.inc(-n)
    end
    def n()
      @0
    end
  end
end
let c1 = counter(4),
    c2 = counter(4) in
  print(c1.inc(4));
  print(c2.dec(2))
end """)
eval(pobj2)

val pdel = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      self.inc(-n)
    end
    def n()
      @0
    end
  end
end
fun dcounter(n)
  object (counter(n))
    def inc(n)
      print(@0.n());
      @0.inc(n)
    end
  end
end
let c1 = counter(4),
    c2 = dcounter(4) in
  print(c1.inc(4));
  print(c2.inc(2))
end """)
eval(pdel)

val pdelor = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(this, n)
      this.inc(-n)
    end
    def n()
      @0
    end
  end
end
fun dcounter(n)
  object (counter(n))
    def inc(n)
      print(@0.n());
      @0.inc(n)
    end
    def dec(n)
      @0.dec(self, n)
    end
  end
end
let c1 = counter(4),
    c2 = dcounter(4) in
  print(c1.inc(4));
  print(c2.dec(2))
end """)
eval(pdelor)

val pher = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      self.inc(-n)
    end
    def n()
      @0
    end
  end
end
fun hcounter(o)
  object () extends o
    def dec(n)
      @0 := @0 - (n * 2)
    end
  end
end
let c1 = counter(4),
    c2 = hcounter(c1) in
  print(c1.inc(4));
  print(c2.dec(2))
end """)
eval(pher)

val phersuper = ProgFun.parse("""
fun counter(n)
  object (n)
    def inc(n)
      @0 := @0 + n
    end
    def dec(n)
      self.inc(-n)
    end
    def n()
      @0
    end
  end
end
fun hcounter(o)
  object () extends o
    def inc(n)
      super.inc(n*2)
    end
  end
end
let c1 = counter(4),
    c2 = hcounter(c1) in
  print(c1.inc(4));
  print(c2.dec(2))
end """)
eval(phersuper)

val plist = ProgFun.parse("""
  fun vazia()
object ()
def imprime() 0 end
def map(f) self end
  end
end

fun cons(h, t)
object (h, t)
def imprime()
print(@0); @1.imprime()
end
def map(f)
cons(f.apply(@0), @1.map(f))
end
end
end
let l1 = cons(1, cons(2, vazia())),
l2 = l1.map(object ()
def apply(o) o * o end
  end) in
l1.imprime(); l2.imprime()
end
""")
eval(plist)

val ptracer = ProgFun.parse("""
fun tracer(f)
object () extends f
def apply(x)
print(x);
super.apply(x)
end
end
end

let fat = object ()
def apply(n)
if n < 2 then
  1
else
  n * self.apply(
    n - 1)
end
end
end, tfat = tracer(fat) in
  print(fat.apply(5));
tfat.apply(5)
end
""")
eval(ptracer)